# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import logging
import threading

_thread_locals = threading.local()

# Create the Logger
logger = logging.getLogger(__package__.rsplit(".", 1)[-1])

# Allow the user to override the logger's handler
if not logger.handlers:
    # Create the Handler for logging data to console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)

    # Create a Formatter for formatting the log messages
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s][XCG][%(name)s]: %(message)s"
    )
    console_handler.setFormatter(formatter)

    # Set the Handler to the Logger
    logger.addHandler(console_handler)

# Allow the user to override the logger's level
if logger.level == logging.NOTSET:
    logger.setLevel(logging.INFO)

logger.propagate = False



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import mimetypes
import operator
import os
import pprint

import magic

from govtech_csg_xcg.securefileupload import logger

from ...conf import settings
from ..helper import add_point_to_guessed_file_type

try:
    import yara
except ImportError:
    yara = None


def perform_yara_matching(file):
    """
    Perform YARA matching.
    """
    if yara is not None:
        logger.debug("[Validation module] - Performing YARA matching")

        yara_dir_path = settings.GLOBAL_UPLOAD_CONFIG["yara_file_location"]
        rules = yara.compile(
            filepaths={
                file_name.split(".")[0]: os.path.join(yara_dir_path, file_name)
                for file_name in os.listdir(yara_dir_path)
                if file_name.lower().endswith(".yar")
                or file_name.lower().endswith(".yara")
            }
        )

        matches = rules.match(data=file.content)

        file.detection_results.yara_matches = matches

    else:
        logger.info(
            "[Validation module] - Skipping YARA validation as YARA library is not installed"
        )


def match_file_signature(file):
    """get the mimetype of the file using magic lib"""
    logger.debug("[Validation module] - Matching file signature")

    return magic.from_buffer(file.content, mime=True)


def get_filename_splits(file_object):
    """split the file by dot and return a list"""
    file_name_splits = list(
        map(lambda x: x.lower(), file_object.basic_information.name.split("."))
    )

    return file_name_splits


def check_file_size_allowed(file, upload_config):
    """
    Check if the file size is within the allowed limits.

    Parameters:
        file (object): An object representing the file to be checked.
        upload_config (dict): A dictionary containing the upload configuration, including the "file_size_limit" key.

    Returns:
        None. The validation result is stored in the "file_size_ok" attribute of the "validation_results" attribute of the "file" object, and the "block" attribute of the "file" object is updated if the file size is not allowed.
    """
    logger.debug("[Validation module] - Validating file size")

    if upload_config["file_size_limit"] is not None:
        file_size_ok = (
            file.basic_information.size / 1000 <= upload_config["file_size_limit"]
        )
    else:
        file_size_ok = True
    file.validation_results.file_size_ok = file_size_ok

    if not file_size_ok:
        file.block = True
        logger.warning("[Validation module] - File size is too big.")


def check_mime_against_whitelist(mime_to_check, upload_config):
    """
    Check if the MIME type of a file is in the allowed whitelist.

    Parameters:
        mime_to_check (str): The MIME type of the file to be checked, in the format "type/subtype".
        upload_config (dict): A dictionary containing the upload configuration, including the "whitelist" key.

    Returns:
        bool: True if the MIME type is in the whitelist, False otherwise.
    """
    return mime_to_check in upload_config["whitelist"]


def check_request_header_mime(file, upload_config):
    """
    Check if the MIME type specified in the request header is in the allowed whitelist.

    Parameters:
        file (object): An object representing the file to be checked.
        upload_config (dict): A dictionary containing the upload configuration, including the "whitelist" key.

    Returns:
        None. The validation result is stored in the "request_whitelist_ok" attribute of the "validation_results" attribute of the "file" object, and the "block" attribute of the "file" object is updated if the MIME type is not in the whitelist.
    """
    logger.debug(
        "[Validation module] - Validating request header MIME type against whitelist"
    )

    mime_whitelist_result = check_mime_against_whitelist(
        file.basic_information.content_type, upload_config
    )

    file.validation_results.request_whitelist_ok = mime_whitelist_result

    if not mime_whitelist_result:
        file.block = True
        logger.warning("[Validation module] - Content-Type not whitelisted")


def check_signature_and_request_mime_match_file_extensions(file):
    """
    Check if the MIME type specified in the request header and the MIME type
    inferred from the file signature match the file extensions.

    Parameters:
        file (object): An object representing the file to be checked. The "extensions"
        attribute of the "detection_results" attribute should contain a list of file
        extensions, and the "content_type" and "signature_mime" attributes of the
        "basic_information" attribute should contain the MIME types specified in the
        request header and inferred from the file signature, respectively.

    Returns:
        None. The validation result is stored in the "matching_extension_signature_request_ok"
        attribute of the "validation_results" attribute of the "file" object, and the
        "mime_manipulation" attribute of the "attack_results" attribute of the "file" object
        is updated if the MIME types do not match the file extensions. The "block" attribute
        of the "file" object is updated if the MIME types do not match.
    """

    extension_matchings = []

    mime_similar_collections = [
        [
            "application/msword",
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            "application/vnd.ms-word",
            "application/rtf",
            "text/rtf",
        ],
        [
            "application/msexcel",
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "application/vnd.ms-excel",
        ],
        [
            "application/mspowerpoint",
            "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            "application/vnd.ms-powerpoint",
        ],
        [
            "application/pdf",
            "application/x-pdf",
            "application/acrobat",
            "applications/vnd.pdf",
            "text/pdf",
            "text/x-pdf",
        ],
        [
            "image/jpeg",
            "image/pjpeg",
            "image/jpg",
            "image/png",
            "image/gif",
            "image/bmp",
            "image/x-windows-bmp",
            "image/x-bitmap",
            "image/x-xbitmap",
            "image/x-win-bitmap",
            "image/x-ms-bmp",
            "image/x-bmp",
        ],
        [
            "video/mp4",
            "video/mpeg",
            "video/quicktime",
            "video/x-msvideo",
            "video/x-ms-wmv",
            "video/x-ms-wm",
            "video/avi",
            "video/msvideo",
            "video/x-ms-asf",
            "video/x-ms-asf-plugin",
        ],
        [
            "audio/mpeg",
            "audio/mp3",
            "audio/x-mpeg",
            "audio/x-mp3",
            "audio/x-mpeg3",
            "audio/mpeg3",
            "audio/mpg",
            "audio/x-mpg",
            "audio/x-mpegaudio",
            "audio/x-mp3-playlist",
        ],
    ]

    for single_file_extension in file.detection_results.extensions:
        file_extension_mime = mimetypes.guess_type("name." + single_file_extension)[0]

        possible_mime_types = [file_extension_mime]
        for mime_similar_collection in mime_similar_collections:
            if file_extension_mime in mime_similar_collection:
                possible_mime_types.extend(mime_similar_collection)
                break

        mime_type_matching = (
            file.basic_information.content_type in possible_mime_types
            and file.detection_results.signature_mime in possible_mime_types
        )

        extension_matchings.append(mime_type_matching)

    all_extensions_match = all(extension_matchings)

    file.validation_results.matching_extension_signature_request_ok = (
        all_extensions_match
    )
    file.attack_results.mime_manipulation = not all_extensions_match

    if not all_extensions_match:
        file.block = True
        logger.warning("[Validation module] - Extension MIME does not match")


def check_file_signature(file, upload_config):
    """
    Check if the MIME type inferred from the file signature is in the allowed whitelist.

    Parameters:
        file (object): An object representing the file to be checked. The "signature_mime"
        attribute of the "detection_results" attribute should contain the MIME type inferred
        from the file signature.
        upload_config (dict): A dictionary containing the upload configuration, including the
        "whitelist" key.

    Returns:
        None. The validation result is stored in the "signature_whitelist_ok" attribute of
        the "validation_results" attribute of the "file" object, and the "block" attribute of
        the "file" object is updated if the MIME type is not in the whitelist.
    """
    logger.debug("[Validation module] - Validating file signature")

    mime_whitelist_result = check_mime_against_whitelist(
        file.detection_results.signature_mime, upload_config
    )
    file.validation_results.signature_whitelist_ok = mime_whitelist_result

    if not mime_whitelist_result:
        file.block = True
        logger.warning("[Validation module] - Signature not whitelisted")


def check_filename_length(file, upload_config):
    """
    Check if the length of the filename is within the allowed limits.

    Parameters:
        file (object): An object representing the file to be checked. The "name"
        attribute of the "basic_information" attribute should contain the filename.
        upload_config (dict): A dictionary containing the upload configuration,
        including the "filename_length_limit" key.

    Returns:
        None. The validation result is stored in the "filename_length_ok" attribute
        of the "validation_results" attribute of the "file" object, and the "block"
        attribute of the "file" object is updated if the filename length is not allowed.
    """
    logger.debug("[Validation module] - Validating filename length")

    if upload_config["filename_length_limit"] is not None:
        length_ok = (
            len(file.basic_information.name) <= upload_config["filename_length_limit"]
        )
    else:
        length_ok = True
    file.validation_results.filename_length_ok = length_ok

    if not length_ok:
        file.block = True
        logger.warning("[Validation module] - Filename length too long")


def check_filename_extensions(file, upload_config):
    """
    Check if all the file extensions in the filename are in the allowed whitelist.

    Parameters:
        file (object): An object representing the file to be checked. The "extensions"
        attribute of the "detection_results" attribute should contain a list of file extensions.
        upload_config (dict): A dictionary containing the upload configuration, including the
        "whitelist" key.

    Returns:
        None. The validation result is stored in the "extensions_whitelist_ok" attribute of
        the "validation_results" attribute of the "file" object, and the "block" attribute of the
        "file" object is updated if any of the file extensions is not in the whitelist.
    """
    logger.debug("[Validation module] - Validating all filename extensions")

    mime_whitelist_results = []
    for single_extension in file.detection_results.extensions:
        curr_file_extension_mime = mimetypes.guess_type("name." + single_extension)[0]
        mime_whitelist_results.append(
            check_mime_against_whitelist(curr_file_extension_mime, upload_config)
        )

    all_extensions_whitelisted = all(mime_whitelist_results)

    file.validation_results.extensions_whitelist_ok = all_extensions_whitelisted
    if not all_extensions_whitelisted:
        file.block = True
        logger.warning("[Validation module] - Extension not whitelisted")

    # TODO: Add detection of alternate media file extensions such as .php5


def check_yara_rules(file):
    """
    Check if the file matches any of the configured YARA rules.

    Parameters:
        file (object): An object representing the file to be checked. The "yara_matches" attribute
        of the "detection_results" attribute should contain a list of YARA match objects.

    Returns:
        None. The validation result is stored in the "yara_rules_ok" attribute of the
        "validation_results" attribute of the "file" object, and the "block" attribute of the
        "file" object is updated if any YARA rule matches. The "block_reason" attribute of the
        "file" object is appended with a message indicating the YARA rule that matched.
    """
    logger.debug("[Validation module] - Validating YARA rules")

    file.validation_results.yara_rules_ok = (
        len(file.detection_results.yara_matches) == 0
    )

    if not file.validation_results.yara_rules_ok:
        file.block = True

        for match in file.detection_results.yara_matches:
            file.append_block_reason("YARA match: " + match.rule)


def check_filename_for_null_byte_injections(file):
    """
    Check if the filename contains any null byte injection attacks.

    Parameters:
        file (object): An object representing the file to be checked. The "filename_splits"
        attribute of the "detection_results" attribute should contain a list of substrings
        of the filename split by dots.

    Returns:
        None. The "null_byte_injection" attribute of the "attack_results" attribute of the
        "file" object is updated if a null byte injection attack is found in the filename.
        The "block" attribute of the "file" object is updated if a null byte inject attack is found.
    """

    logger.debug("[Validation module] - Validating for null byte injections")

    null_byte_found = False

    for file_name_split in file.detection_results.filename_splits:
        null_byte_found_in_split = (
            "0x00" in file_name_split
            or "%00" in file_name_split
            or "\0" in file_name_split
        )

        if null_byte_found_in_split:
            null_byte_found = True
            break

    file.attack_results.null_byte_injection = null_byte_found
    if null_byte_found:
        file.block = True
        logger.warning("[Validation module] - Null byte injection found")


def guess_mime_type(file):
    """
    Guess the MIME type of the file by considering its file signature, file extension, and
    Content-Type header.

    Parameters:
        file (object): An object representing the file to be checked. The "signature_mime"
        attribute of the "detection_results" attribute, the "extensions" attribute of the
        "detection_results" attribute, and the "content_type" attribute of the "basic_information"
        attribute should contain relevant information for guessing the MIME type.

    Returns:
        None. The "guessed_mime" attribute of the "detection_results" attribute of the "file"
        object is updated with the guessed MIME type.
    """
    logger.debug("[Validation module] - Guessing MIME type")

    guessing_scores = {mime_type: 0 for mime_type in list(mimetypes.types_map.values())}
    total_points_given = 0
    total_points_overall = 0

    # Adding file signature information
    file_signature_mime = file.detection_results.signature_mime
    if file_signature_mime in guessing_scores.keys():
        guessing_scores[file_signature_mime] += 1
        total_points_given += 1
    total_points_overall += 1

    # Adding file extension information
    main_file_extension = file.detection_results.extensions[0]
    main_mime_type = mimetypes.guess_type("name." + main_file_extension)[0]
    if main_mime_type in guessing_scores.keys():
        guessing_scores[main_mime_type] += 1
        total_points_given += 1
    total_points_overall += 1

    # Adding Content-Type header information
    content_type_mime = file.basic_information.content_type
    if content_type_mime in guessing_scores.keys():
        guessing_scores[content_type_mime] += 1
        total_points_given += 1
    total_points_overall += 1

    # Evaluating maliciousness
    sorted_guessing_scores = {
        k: v
        for k, v in sorted(guessing_scores.items(), key=lambda item: item[1])
        if v > 0
    }
    logger.debug(f"[Validation module] - {pprint.pformat(sorted_guessing_scores)}")
    logger.debug(
        f"[Validation module] - {total_points_overall=} - {total_points_given=}"
    )

    guessed_mime_type = max(guessing_scores.items(), key=operator.itemgetter(1))[0]
    file.detection_results.guessed_mime = guessed_mime_type


def validate_file(file, upload_config):
    """
    Validate the file to ensure it meets the requirements specified in the
    upload config.

    Args:
        file: A File object containing the file to be validated.
        upload_config: A dictionary containing the configuration for file uploads.

    Returns:
        None. Modifies the file object in place.
    """
    logger.debug(
        f"{file.basic_information.name} - [Validation module] - Starting basic detection"
    )

    # Retrieve basic file information
    filename_splits = get_filename_splits(file)
    file.detection_results.filename_splits = filename_splits
    file.detection_results.extensions = [filename_splits[-1]]
    main_file_extension = file.detection_results.extensions[0]
    main_extension_mime_type = mimetypes.guess_type("name." + main_file_extension)[0]
    add_point_to_guessed_file_type(file, main_extension_mime_type)

    # Detecting file signature
    signature_mime = match_file_signature(file)
    file.detection_results.signature_mime = signature_mime
    add_point_to_guessed_file_type(file, signature_mime)

    # Validate file information
    check_file_size_allowed(file, upload_config)

    check_request_header_mime(file, upload_config)

    check_signature_and_request_mime_match_file_extensions(file)

    check_file_signature(file, upload_config)

    check_filename_length(file, upload_config)

    check_filename_extensions(file, upload_config)

    check_filename_for_null_byte_injections(file)

    if not file.block:
        # Match YARA rules
        perform_yara_matching(file)
        check_yara_rules(file)

    if not file.block:
        guess_mime_type(file)



#!/usr/bin/env python

# flake8: noqa

## PDFID was adjusted to let it generate a JSON supported report

__description__ = "Tool to test a PDF file"
__author__ = "Didier Stevens"
__version__ = "0.2.7"
__date__ = "2019/11/05"

"""

Tool to test a PDF file

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2009/03/27: start
  2009/03/28: scan option
  2009/03/29: V0.0.2: xml output
  2009/03/31: V0.0.3: /ObjStm suggested by Dion
  2009/04/02: V0.0.4: added ErrorMessage
  2009/04/20: V0.0.5: added Dates
  2009/04/21: V0.0.6: added entropy
  2009/04/22: added disarm
  2009/04/29: finished disarm
  2009/05/13: V0.0.7: added cPDFEOF
  2009/07/24: V0.0.8: added /AcroForm and /RichMedia, simplified %PDF header regex, extra date format (without TZ)
  2009/07/25: added input redirection, option --force
  2009/10/13: V0.0.9: added detection for CVE-2009-3459; added /RichMedia to disarm
  2010/01/11: V0.0.10: relaxed %PDF header checking
  2010/04/28: V0.0.11: added /Launch
  2010/09/21: V0.0.12: fixed cntCharsAfterLastEOF bug; fix by Russell Holloway
  2011/12/29: updated for Python 3, added keyword /EmbeddedFile
  2012/03/03: added PDFiD2JSON; coded by Brandon Dixon
  2013/02/10: V0.1.0: added http/https support; added support for ZIP file with password 'infected'
  2013/03/11: V0.1.1: fixes for Python 3
  2013/03/13: V0.1.2: Added error handling for files; added /XFA
  2013/11/01: V0.2.0: Added @file & plugins
  2013/11/02: continue
  2013/11/04: added options -c, -m, -v
  2013/11/06: added option -S
  2013/11/08: continue
  2013/11/09: added option -o
  2013/11/15: refactoring
  2014/09/30: added CSV header
  2014/10/16: V0.2.1: added output when plugin & file not pdf
  2014/10/18: some fixes for Python 3
  2015/08/12: V0.2.2: added option pluginoptions
  2015/08/13: added plugin Instructions method
  2016/04/12: added option literal
  2017/10/29: added pdfid.ini support
  2017/11/05: V0.2.3: added option -n
  2018/01/03: V0.2.4: bugfix entropy calculation for PDFs without streams; sample 28cb208d976466b295ee879d2d233c8a https://twitter.com/DubinRan/status/947783629123416069
  2018/01/15: bugfix ConfigParser privately reported
  2018/01/29: bugfix oPDFEOF.cntCharsAfterLastEOF when no %%EOF
  2018/07/05: V0.2.5 introduced cExpandFilenameArguments; renamed option literal to literalfilenames
  2019/09/30: V0.2.6 color bugfix, thanks to Leo
  2019/11/05: V0.2.7 fixed plugin path when compiled with pyinstaller

Todo:
  - update XML example (entropy, EOF)
  - code review, cleanup
"""

import collections
import fnmatch
import glob
import io
import json
import math
import operator
import optparse
import os
import os.path
import sys
import traceback
import xml.dom.minidom
import zipfile

from . import logger

if sys.version_info[0] >= 3:
    import urllib.request as urllib23
else:
    import urllib2 as urllib23
if sys.version_info[0] >= 3:
    import configparser as ConfigParser
else:
    import ConfigParser


# Convert 2 Bytes If Python 3
def C2BIP3(string):
    if sys.version_info[0] > 2:
        return bytes([ord(x) for x in string])
    else:
        return string


class cBinaryFile:
    def __init__(self, file):
        self.file = file
        if file == "":
            self.infile = sys.stdin
        elif isinstance(file, bytes):
            self.infile = io.BytesIO(file)
        elif file.lower().startswith("http://") or file.lower().startswith("https://"):
            try:
                if sys.hexversion >= 0x020601F0:
                    self.infile = urllib23.urlopen(file, timeout=5)
                else:
                    self.infile = urllib23.urlopen(file)
            except urllib23.HTTPError:
                logger.error("Error accessing URL %s" % file)
                logger.info(sys.exc_info()[1])
                sys.exit()
        elif file.lower().endswith(".zip"):
            try:
                self.zipfile = zipfile.ZipFile(file, "r")
                self.infile = self.zipfile.open(
                    self.zipfile.infolist()[0], "r", C2BIP3("infected")
                )
            except:
                logger.error("Error opening file %s" % file)
                logger.info(sys.exc_info()[1])
                sys.exit()
        else:
            try:
                self.infile = open(file, "rb")
            except:
                logger.error("Error opening file %s" % file)
                logger.info(sys.exc_info()[1])
                sys.exit()
        self.ungetted = []

    def byte(self):
        if len(self.ungetted) != 0:
            return self.ungetted.pop()
        inbyte = self.infile.read(1)
        if not inbyte or inbyte == "":
            self.infile.close()
            return None
        return ord(inbyte)

    def bytes(self, size):
        if size <= len(self.ungetted):
            result = self.ungetted[0:size]
            del self.ungetted[0:size]
            return result
        inbytes = self.infile.read(size - len(self.ungetted))
        if inbytes == "":
            self.infile.close()
        if type(inbytes) == str:
            result = self.ungetted + [ord(b) for b in inbytes]
        else:
            result = self.ungetted + [b for b in inbytes]
        self.ungetted = []
        return result

    def unget(self, byte):
        self.ungetted.append(byte)

    def ungets(self, bytes):
        bytes.reverse()
        self.ungetted.extend(bytes)


class cPDFDate:
    def __init__(self):
        self.state = 0

    def parse(self, char):
        if char == "D":
            self.state = 1
            return None
        elif self.state == 1:
            if char == ":":
                self.state = 2
                self.digits1 = ""
            else:
                self.state = 0
            return None
        elif self.state == 2:
            if len(self.digits1) < 14:
                if char >= "0" and char <= "9":
                    self.digits1 += char
                    return None
                else:
                    self.state = 0
                    return None
            elif char == "+" or char == "-" or char == "Z":
                self.state = 3
                self.digits2 = ""
                self.TZ = char
                return None
            elif char == '"':
                self.state = 0
                self.date = "D:" + self.digits1
                return self.date
            elif char < "0" or char > "9":
                self.state = 0
                self.date = "D:" + self.digits1
                return self.date
            else:
                self.state = 0
                return None
        elif self.state == 3:
            if len(self.digits2) < 2:
                if char >= "0" and char <= "9":
                    self.digits2 += char
                    return None
                else:
                    self.state = 0
                    return None
            elif len(self.digits2) == 2:
                if char == "'":
                    self.digits2 += char
                    return None
                else:
                    self.state = 0
                    return None
            elif len(self.digits2) < 5:
                if char >= "0" and char <= "9":
                    self.digits2 += char
                    if len(self.digits2) == 5:
                        self.state = 0
                        self.date = "D:" + self.digits1 + self.TZ + self.digits2
                        return self.date
                    else:
                        return None
                else:
                    self.state = 0
                    return None


def fEntropy(countByte, countTotal):
    x = float(countByte) / countTotal
    if x > 0:
        return -x * math.log(x, 2)
    else:
        return 0.0


class cEntropy:
    def __init__(self):
        self.allBucket = [0 for i in range(0, 256)]
        self.streamBucket = [0 for i in range(0, 256)]

    def add(self, byte, insideStream):
        self.allBucket[byte] += 1
        if insideStream:
            self.streamBucket[byte] += 1

    def removeInsideStream(self, byte):
        if self.streamBucket[byte] > 0:
            self.streamBucket[byte] -= 1

    def calc(self):
        self.nonStreamBucket = map(operator.sub, self.allBucket, self.streamBucket)
        allCount = sum(self.allBucket)
        streamCount = sum(self.streamBucket)
        nonStreamCount = sum(self.nonStreamBucket)
        if streamCount == 0:
            return (
                allCount,
                sum(map(lambda x: fEntropy(x, allCount), self.allBucket)),
                streamCount,
                None,
                nonStreamCount,
                sum(map(lambda x: fEntropy(x, nonStreamCount), self.nonStreamBucket)),
            )
        else:
            return (
                allCount,
                sum(map(lambda x: fEntropy(x, allCount), self.allBucket)),
                streamCount,
                sum(map(lambda x: fEntropy(x, streamCount), self.streamBucket)),
                nonStreamCount,
                sum(map(lambda x: fEntropy(x, nonStreamCount), self.nonStreamBucket)),
            )


class cPDFEOF:
    def __init__(self):
        self.token = ""
        self.cntEOFs = 0

    def parse(self, char):
        if self.cntEOFs > 0:
            self.cntCharsAfterLastEOF += 1
        if self.token == "" and char == "%":
            self.token += char
            return
        elif self.token == "%" and char == "%":
            self.token += char
            return
        elif self.token == "%%" and char == "E":
            self.token += char
            return
        elif self.token == "%%E" and char == "O":
            self.token += char
            return
        elif self.token == "%%EO" and char == "F":
            self.token += char
            return
        elif self.token == "%%EOF" and (
            char == "\n" or char == "\r" or char == " " or char == "\t"
        ):
            self.cntEOFs += 1
            self.cntCharsAfterLastEOF = 0
            if char == "\n":
                self.token = ""
            else:
                self.token += char
            return
        elif self.token == "%%EOF\r":
            if char == "\n":
                self.cntCharsAfterLastEOF = 0
            self.token = ""
        else:
            self.token = ""


def FindPDFHeaderRelaxed(oBinaryFile):
    bytes = oBinaryFile.bytes(1024)
    index = "".join([chr(byte) for byte in bytes]).find("%PDF")
    if index == -1:
        oBinaryFile.ungets(bytes)
        return ([], None)
    for endHeader in range(index + 4, index + 4 + 10):
        if bytes[endHeader] == 10 or bytes[endHeader] == 13:
            break
    oBinaryFile.ungets(bytes[endHeader:])
    return (bytes[0:endHeader], "".join([chr(byte) for byte in bytes[index:endHeader]]))


def Hexcode2String(char):
    if type(char) == int:
        return "#%02x" % char
    else:
        return char


def SwapCase(char):
    if type(char) == int:
        return ord(chr(char).swapcase())
    else:
        return char.swapcase()


def HexcodeName2String(hexcodeName):
    return "".join(map(Hexcode2String, hexcodeName))


def SwapName(wordExact):
    return map(SwapCase, wordExact)


def UpdateWords(
    word,
    wordExact,
    slash,
    words,
    hexcode,
    allNames,
    lastName,
    insideStream,
    oEntropy,
    disarmed_buffer=None,
):
    if word != "":
        if slash + word in words:
            words[slash + word][0] += 1
            if hexcode:
                words[slash + word][1] += 1
        elif slash == "/" and allNames:
            words[slash + word] = [1, 0]
            if hexcode:
                words[slash + word][1] += 1
        if slash == "/":
            lastName = slash + word
        if slash == "":
            if word == "stream":
                insideStream = True
            if word == "endstream":
                if insideStream == True and oEntropy != None:
                    for char in "endstream":
                        oEntropy.removeInsideStream(ord(char))
                insideStream = False
        if disarmed_buffer is not None:
            if slash == "/" and "/" + word in (
                "/JS",
                "/JavaScript",
                "/AA",
                "/OpenAction",
                "/JBIG2Decode",
                "/RichMedia",
                "/Launch",
            ):
                key_word_out = HexcodeName2String(SwapName(wordExact))
            else:
                key_word_out = HexcodeName2String(wordExact)
            disarmed_buffer.write(C2BIP3(key_word_out))
    return ("", [], False, lastName, insideStream)


class cCVE_2009_3459:
    def __init__(self):
        self.count = 0

    def Check(self, lastName, word):
        if (
            lastName == "/Colors" and word.isdigit() and int(word) > 2**24
        ):  # decided to alert when the number of colors is expressed with more than 3 bytes
            self.count += 1


def XMLAddAttribute(xmlDoc, name, value=None):
    att = xmlDoc.createAttribute(name)
    xmlDoc.documentElement.setAttributeNode(att)
    if value != None:
        att.nodeValue = value
    return att


def GetScriptPath():
    if getattr(sys, "frozen", False):
        return os.path.dirname(sys.executable)
    else:
        return os.path.dirname(sys.argv[0])


def ParseINIFile():
    oConfigParser = ConfigParser.ConfigParser(allow_no_value=True)
    oConfigParser.optionxform = str
    oConfigParser.read(os.path.join(GetScriptPath(), "pdfid.ini"))
    keywords = []
    if oConfigParser.has_section("keywords"):
        for key, value in oConfigParser.items("keywords"):
            if not key in keywords:
                keywords.append(key)
    return keywords


def PDFiD(
    file,
    allNames=False,
    extraData=False,
    disarm=False,
    force=False,
    disarmed_filebuffers=None,
    filebuffer=None,
):
    """Example of XML output:
    <PDFiD ErrorOccured="False" ErrorMessage="" Filename="test.pdf" Header="%PDF-1.1" IsPDF="True" Version="0.0.4" Entropy="4.28">
            <Keywords>
                    <Keyword Count="7" HexcodeCount="0" Name="obj"/>
                    <Keyword Count="7" HexcodeCount="0" Name="endobj"/>
                    <Keyword Count="1" HexcodeCount="0" Name="stream"/>
                    <Keyword Count="1" HexcodeCount="0" Name="endstream"/>
                    <Keyword Count="1" HexcodeCount="0" Name="xref"/>
                    <Keyword Count="1" HexcodeCount="0" Name="trailer"/>
                    <Keyword Count="1" HexcodeCount="0" Name="startxref"/>
                    <Keyword Count="1" HexcodeCount="0" Name="/Page"/>
                    <Keyword Count="0" HexcodeCount="0" Name="/Encrypt"/>
                    <Keyword Count="1" HexcodeCount="0" Name="/JS"/>
                    <Keyword Count="1" HexcodeCount="0" Name="/JavaScript"/>
                    <Keyword Count="0" HexcodeCount="0" Name="/AA"/>
                    <Keyword Count="1" HexcodeCount="0" Name="/OpenAction"/>
                    <Keyword Count="0" HexcodeCount="0" Name="/JBIG2Decode"/>
            </Keywords>
            <Dates>
                    <Date Value="D:20090128132916+01'00" Name="/ModDate"/>
            </Dates>
    </PDFiD>
    """
    if disarmed_filebuffers is None:
        disarmed_filebuffers = []

    word = ""
    wordExact = []
    hexcode = False
    lastName = ""
    insideStream = False
    keywords = [
        "obj",
        "endobj",
        "stream",
        "endstream",
        "xref",
        "trailer",
        "startxref",
        "/Page",
        "/Encrypt",
        "/ObjStm",
        "/JS",
        "/JavaScript",
        "/AA",
        "/OpenAction",
        "/AcroForm",
        "/JBIG2Decode",
        "/RichMedia",
        "/Launch",
        "/EmbeddedFile",
        "/XFA",
    ]
    words = {}
    dates = []
    for extrakeyword in ParseINIFile():
        if not extrakeyword in keywords:
            keywords.append(extrakeyword)
    for keyword in keywords:
        words[keyword] = [0, 0]
    slash = ""
    xmlDoc = xml.dom.minidom.getDOMImplementation().createDocument(None, "PDFiD", None)
    XMLAddAttribute(xmlDoc, "Version", __version__)
    XMLAddAttribute(xmlDoc, "Filename", file)
    attErrorOccured = XMLAddAttribute(xmlDoc, "ErrorOccured", "False")
    attErrorMessage = XMLAddAttribute(xmlDoc, "ErrorMessage", "")

    oPDFDate = None
    oEntropy = None
    oPDFEOF = None
    oCVE_2009_3459 = cCVE_2009_3459()
    try:
        attIsPDF = xmlDoc.createAttribute("IsPDF")
        xmlDoc.documentElement.setAttributeNode(attIsPDF)
        if filebuffer is not None:
            oBinaryFile = cBinaryFile(filebuffer)
        else:
            oBinaryFile = cBinaryFile(file)
        if extraData:
            oPDFDate = cPDFDate()
            oEntropy = cEntropy()
            oPDFEOF = cPDFEOF()
        (bytesHeader, pdfHeader) = FindPDFHeaderRelaxed(oBinaryFile)
        if disarm:
            disarmed_pdf_buffer = io.BytesIO()
            for byteHeader in bytesHeader:
                disarmed_pdf_buffer.write(C2BIP3(chr(byteHeader)))
        else:
            disarmed_pdf_buffer = None
        if oEntropy != None:
            for byteHeader in bytesHeader:
                oEntropy.add(byteHeader, insideStream)
        if pdfHeader == None and not force:
            attIsPDF.nodeValue = "False"
            return xmlDoc
        else:
            if pdfHeader == None:
                attIsPDF.nodeValue = "False"
                pdfHeader = ""
            else:
                attIsPDF.nodeValue = "True"
            att = xmlDoc.createAttribute("Header")
            att.nodeValue = repr(pdfHeader[0:10]).strip("'")
            xmlDoc.documentElement.setAttributeNode(att)
        byte = oBinaryFile.byte()
        while byte != None:
            char = chr(byte)
            charUpper = char.upper()
            if (
                charUpper >= "A"
                and charUpper <= "Z"
                or charUpper >= "0"
                and charUpper <= "9"
            ):
                word += char
                wordExact.append(char)
            elif slash == "/" and char == "#":
                d1 = oBinaryFile.byte()
                if d1 != None:
                    d2 = oBinaryFile.byte()
                    if (
                        d2 != None
                        and (
                            chr(d1) >= "0"
                            and chr(d1) <= "9"
                            or chr(d1).upper() >= "A"
                            and chr(d1).upper() <= "F"
                        )
                        and (
                            chr(d2) >= "0"
                            and chr(d2) <= "9"
                            or chr(d2).upper() >= "A"
                            and chr(d2).upper() <= "F"
                        )
                    ):
                        word += chr(int(chr(d1) + chr(d2), 16))
                        wordExact.append(int(chr(d1) + chr(d2), 16))
                        hexcode = True
                        if oEntropy != None:
                            oEntropy.add(d1, insideStream)
                            oEntropy.add(d2, insideStream)
                        if oPDFEOF != None:
                            oPDFEOF.parse(d1)
                            oPDFEOF.parse(d2)
                    else:
                        oBinaryFile.unget(d2)
                        oBinaryFile.unget(d1)
                        (
                            word,
                            wordExact,
                            hexcode,
                            lastName,
                            insideStream,
                        ) = UpdateWords(
                            word,
                            wordExact,
                            slash,
                            words,
                            hexcode,
                            allNames,
                            lastName,
                            insideStream,
                            oEntropy,
                            disarmed_pdf_buffer,
                        )
                        if disarm:
                            disarmed_pdf_buffer.write(C2BIP3(char))
                else:
                    oBinaryFile.unget(d1)
                    (word, wordExact, hexcode, lastName, insideStream) = UpdateWords(
                        word,
                        wordExact,
                        slash,
                        words,
                        hexcode,
                        allNames,
                        lastName,
                        insideStream,
                        oEntropy,
                        disarmed_pdf_buffer,
                    )
                    if disarm:
                        disarmed_pdf_buffer.write(C2BIP3(char))
            else:
                oCVE_2009_3459.Check(lastName, word)

                (word, wordExact, hexcode, lastName, insideStream) = UpdateWords(
                    word,
                    wordExact,
                    slash,
                    words,
                    hexcode,
                    allNames,
                    lastName,
                    insideStream,
                    oEntropy,
                    disarmed_pdf_buffer,
                )
                if char == "/":
                    slash = "/"
                else:
                    slash = ""
                if disarm:
                    disarmed_pdf_buffer.write(C2BIP3(char))

            if oPDFDate != None and oPDFDate.parse(char) != None:
                dates.append([oPDFDate.date, lastName])

            if oEntropy != None:
                oEntropy.add(byte, insideStream)

            if oPDFEOF != None:
                oPDFEOF.parse(char)

            byte = oBinaryFile.byte()
        (word, wordExact, hexcode, lastName, insideStream) = UpdateWords(
            word,
            wordExact,
            slash,
            words,
            hexcode,
            allNames,
            lastName,
            insideStream,
            oEntropy,
            disarmed_pdf_buffer,
        )

        # check to see if file ended with %%EOF.  If so, we can reset charsAfterLastEOF and add one to EOF count.  This is never performed in
        # the parse function because it never gets called due to hitting the end of file.
        if byte == None and oPDFEOF != None:
            if oPDFEOF.token == "%%EOF":
                oPDFEOF.cntEOFs += 1
                oPDFEOF.cntCharsAfterLastEOF = 0
                oPDFEOF.token = ""

    except SystemExit:
        sys.exit()
    except:
        attErrorOccured.nodeValue = "True"
        attErrorMessage.nodeValue = traceback.format_exc()

    if disarm:
        disarmed_filebuffers.append(disarmed_pdf_buffer.getvalue())
        disarmed_pdf_buffer.close()

    attEntropyAll = xmlDoc.createAttribute("TotalEntropy")
    xmlDoc.documentElement.setAttributeNode(attEntropyAll)
    attCountAll = xmlDoc.createAttribute("TotalCount")
    xmlDoc.documentElement.setAttributeNode(attCountAll)
    attEntropyStream = xmlDoc.createAttribute("StreamEntropy")
    xmlDoc.documentElement.setAttributeNode(attEntropyStream)
    attCountStream = xmlDoc.createAttribute("StreamCount")
    xmlDoc.documentElement.setAttributeNode(attCountStream)
    attEntropyNonStream = xmlDoc.createAttribute("NonStreamEntropy")
    xmlDoc.documentElement.setAttributeNode(attEntropyNonStream)
    attCountNonStream = xmlDoc.createAttribute("NonStreamCount")
    xmlDoc.documentElement.setAttributeNode(attCountNonStream)
    if oEntropy != None:
        (
            countAll,
            entropyAll,
            countStream,
            entropyStream,
            countNonStream,
            entropyNonStream,
        ) = oEntropy.calc()
        attEntropyAll.nodeValue = "%f" % entropyAll
        attCountAll.nodeValue = "%d" % countAll
        if entropyStream == None:
            attEntropyStream.nodeValue = "N/A     "
        else:
            attEntropyStream.nodeValue = "%f" % entropyStream
        attCountStream.nodeValue = "%d" % countStream
        attEntropyNonStream.nodeValue = "%f" % entropyNonStream
        attCountNonStream.nodeValue = "%d" % countNonStream
    else:
        attEntropyAll.nodeValue = ""
        attCountAll.nodeValue = ""
        attEntropyStream.nodeValue = ""
        attCountStream.nodeValue = ""
        attEntropyNonStream.nodeValue = ""
        attCountNonStream.nodeValue = ""
    attCountEOF = xmlDoc.createAttribute("CountEOF")
    xmlDoc.documentElement.setAttributeNode(attCountEOF)
    attCountCharsAfterLastEOF = xmlDoc.createAttribute("CountCharsAfterLastEOF")
    xmlDoc.documentElement.setAttributeNode(attCountCharsAfterLastEOF)
    if oPDFEOF != None:
        attCountEOF.nodeValue = "%d" % oPDFEOF.cntEOFs
        if oPDFEOF.cntEOFs > 0:
            attCountCharsAfterLastEOF.nodeValue = "%d" % oPDFEOF.cntCharsAfterLastEOF
        else:
            attCountCharsAfterLastEOF.nodeValue = ""
    else:
        attCountEOF.nodeValue = ""
        attCountCharsAfterLastEOF.nodeValue = ""

    eleKeywords = xmlDoc.createElement("Keywords")
    xmlDoc.documentElement.appendChild(eleKeywords)
    for keyword in keywords:
        eleKeyword = xmlDoc.createElement("Keyword")
        eleKeywords.appendChild(eleKeyword)
        att = xmlDoc.createAttribute("Name")
        att.nodeValue = keyword
        eleKeyword.setAttributeNode(att)
        att = xmlDoc.createAttribute("Count")
        att.nodeValue = str(words[keyword][0])
        eleKeyword.setAttributeNode(att)
        att = xmlDoc.createAttribute("HexcodeCount")
        att.nodeValue = str(words[keyword][1])
        eleKeyword.setAttributeNode(att)
    eleKeyword = xmlDoc.createElement("Keyword")
    eleKeywords.appendChild(eleKeyword)
    att = xmlDoc.createAttribute("Name")
    att.nodeValue = "/Colors > 2^24"
    eleKeyword.setAttributeNode(att)
    att = xmlDoc.createAttribute("Count")
    att.nodeValue = str(oCVE_2009_3459.count)
    eleKeyword.setAttributeNode(att)
    att = xmlDoc.createAttribute("HexcodeCount")
    att.nodeValue = str(0)
    eleKeyword.setAttributeNode(att)
    if allNames:
        keys = sorted(words.keys())
        for word in keys:
            if not word in keywords:
                eleKeyword = xmlDoc.createElement("Keyword")
                eleKeywords.appendChild(eleKeyword)
                att = xmlDoc.createAttribute("Name")
                att.nodeValue = word
                eleKeyword.setAttributeNode(att)
                att = xmlDoc.createAttribute("Count")
                att.nodeValue = str(words[word][0])
                eleKeyword.setAttributeNode(att)
                att = xmlDoc.createAttribute("HexcodeCount")
                att.nodeValue = str(words[word][1])
                eleKeyword.setAttributeNode(att)
    eleDates = xmlDoc.createElement("Dates")
    xmlDoc.documentElement.appendChild(eleDates)
    dates.sort(key=lambda x: x[0])
    for date in dates:
        eleDate = xmlDoc.createElement("Date")
        eleDates.appendChild(eleDate)
        att = xmlDoc.createAttribute("Value")
        att.nodeValue = date[0]
        eleDate.setAttributeNode(att)
        att = xmlDoc.createAttribute("Name")
        att.nodeValue = date[1]
        eleDate.setAttributeNode(att)
    return xmlDoc


def PDFID2Dict(xmlDoc, nozero, force, list_of_dict):
    filename_dict = {}
    filename_dict["version"] = xmlDoc.documentElement.getAttribute("Version")
    filename_dict["filename"] = xmlDoc.documentElement.getAttribute("Filename")
    if xmlDoc.documentElement.getAttribute("ErrorOccured") == "True":
        filename_dict["error_occured"] = xmlDoc.documentElement.getAttribute(
            "ErrorMessage"
        )
        return
    if not force and xmlDoc.documentElement.getAttribute("IsPDF") == "False":
        filename_dict["error_occured"] = " Not a PDF document\n"
        return
    filename_dict["header"] = xmlDoc.documentElement.getAttribute("Header")
    for node in xmlDoc.documentElement.getElementsByTagName("Keywords")[0].childNodes:
        if not nozero or nozero and int(node.getAttribute("Count")) > 0:
            filename_dict[node.getAttribute("Name")] = int(node.getAttribute("Count"))
    if xmlDoc.documentElement.getAttribute("CountEOF") != "":
        filename_dict["eof"] = int(xmlDoc.documentElement.getAttribute("CountEOF"))
    if xmlDoc.documentElement.getAttribute("CountCharsAfterLastEOF") != "":
        filename_dict["after_last_eof"] = int(
            xmlDoc.documentElement.getAttribute("CountCharsAfterLastEOF")
        )
    for node in xmlDoc.documentElement.getElementsByTagName("Dates")[0].childNodes:
        filename_dict[node.getAttribute("Value")] = node.getAttribute("Name")
    if xmlDoc.documentElement.getAttribute("TotalEntropy") != "":
        filename_dict["entropy"] = {
            "total": xmlDoc.documentElement.getAttribute("TotalEntropy"),
            "bytes": "%10s bytes" % xmlDoc.documentElement.getAttribute("TotalCount"),
        }
    if xmlDoc.documentElement.getAttribute("StreamEntropy") != "":
        filename_dict["entropy_inside_streams"] = {
            "total": xmlDoc.documentElement.getAttribute("StreamEntropy"),
            "bytes": "%10s bytes" % xmlDoc.documentElement.getAttribute("StreamCount"),
        }
    if xmlDoc.documentElement.getAttribute("NonStreamEntropy") != "":
        filename_dict["entropy_outside_streams"] = {
            "total": xmlDoc.documentElement.getAttribute("NonStreamEntropy"),
            "bytes": "%10s bytes"
            % xmlDoc.documentElement.getAttribute("NonStreamCount"),
        }
    list_of_dict.append(filename_dict)


def PDFiD2String(xmlDoc, nozero, force):
    result = "PDFiD {} {}\n".format(
        xmlDoc.documentElement.getAttribute("Version"),
        xmlDoc.documentElement.getAttribute("Filename"),
    )
    if xmlDoc.documentElement.getAttribute("ErrorOccured") == "True":
        return (
            result
            + "***Error occured***\n%s\n"
            % xmlDoc.documentElement.getAttribute("ErrorMessage")
        )
    if not force and xmlDoc.documentElement.getAttribute("IsPDF") == "False":
        return result + " Not a PDF document\n"
    result += " PDF Header: %s\n" % xmlDoc.documentElement.getAttribute("Header")
    for node in xmlDoc.documentElement.getElementsByTagName("Keywords")[0].childNodes:
        if not nozero or nozero and int(node.getAttribute("Count")) > 0:
            result += " %-16s %7d" % (
                node.getAttribute("Name"),
                int(node.getAttribute("Count")),
            )
            if int(node.getAttribute("HexcodeCount")) > 0:
                result += "(%d)" % int(node.getAttribute("HexcodeCount"))
            result += "\n"
    if xmlDoc.documentElement.getAttribute("CountEOF") != "":
        result += " %-16s %7d\n" % (
            "%%EOF",
            int(xmlDoc.documentElement.getAttribute("CountEOF")),
        )
    if xmlDoc.documentElement.getAttribute("CountCharsAfterLastEOF") != "":
        result += " %-16s %7d\n" % (
            "After last %%EOF",
            int(xmlDoc.documentElement.getAttribute("CountCharsAfterLastEOF")),
        )
    for node in xmlDoc.documentElement.getElementsByTagName("Dates")[0].childNodes:
        result += " %-23s %s\n" % (
            node.getAttribute("Value"),
            node.getAttribute("Name"),
        )
    if xmlDoc.documentElement.getAttribute("TotalEntropy") != "":
        result += " Total entropy:           %s (%10s bytes)\n" % (
            xmlDoc.documentElement.getAttribute("TotalEntropy"),
            xmlDoc.documentElement.getAttribute("TotalCount"),
        )
    if xmlDoc.documentElement.getAttribute("StreamEntropy") != "":
        result += " Entropy inside streams:  %s (%10s bytes)\n" % (
            xmlDoc.documentElement.getAttribute("StreamEntropy"),
            xmlDoc.documentElement.getAttribute("StreamCount"),
        )
    if xmlDoc.documentElement.getAttribute("NonStreamEntropy") != "":
        result += " Entropy outside streams: %s (%10s bytes)\n" % (
            xmlDoc.documentElement.getAttribute("NonStreamEntropy"),
            xmlDoc.documentElement.getAttribute("NonStreamCount"),
        )
    return result


class cCount:
    def __init__(self, count, hexcode):
        self.count = count
        self.hexcode = hexcode


class cPDFiD:
    def __init__(self, xmlDoc, force):
        self.version = xmlDoc.documentElement.getAttribute("Version")
        self.filename = xmlDoc.documentElement.getAttribute("Filename")
        self.errorOccured = (
            xmlDoc.documentElement.getAttribute("ErrorOccured") == "True"
        )
        self.errorMessage = xmlDoc.documentElement.getAttribute("ErrorMessage")
        self.isPDF = None
        if self.errorOccured:
            return
        self.isPDF = xmlDoc.documentElement.getAttribute("IsPDF") == "True"
        if not force and not self.isPDF:
            return
        self.header = xmlDoc.documentElement.getAttribute("Header")
        self.keywords = {}
        for node in xmlDoc.documentElement.getElementsByTagName("Keywords")[
            0
        ].childNodes:
            self.keywords[node.getAttribute("Name")] = cCount(
                int(node.getAttribute("Count")), int(node.getAttribute("HexcodeCount"))
            )
        self.obj = self.keywords["obj"]
        self.endobj = self.keywords["endobj"]
        self.stream = self.keywords["stream"]
        self.endstream = self.keywords["endstream"]
        self.xref = self.keywords["xref"]
        self.trailer = self.keywords["trailer"]
        self.startxref = self.keywords["startxref"]
        self.page = self.keywords["/Page"]
        self.encrypt = self.keywords["/Encrypt"]
        self.objstm = self.keywords["/ObjStm"]
        self.js = self.keywords["/JS"]
        self.javascript = self.keywords["/JavaScript"]
        self.aa = self.keywords["/AA"]
        self.openaction = self.keywords["/OpenAction"]
        self.acroform = self.keywords["/AcroForm"]
        self.jbig2decode = self.keywords["/JBIG2Decode"]
        self.richmedia = self.keywords["/RichMedia"]
        self.launch = self.keywords["/Launch"]
        self.embeddedfile = self.keywords["/EmbeddedFile"]
        self.xfa = self.keywords["/XFA"]
        self.colors_gt_2_24 = self.keywords["/Colors > 2^24"]


def Print(lines, options):
    if not options.json:
        logger.debug(lines)
        filename = None
        if options.scan:
            filename = "PDFiD.log"
        if options.output != "":
            filename = options.output
        if filename:
            logfile = open(filename, "a")
            logfile.write(lines + "\n")
            logfile.close()


def Quote(value, separator, quote):
    if isinstance(value, str):
        if separator in value:
            return quote + value + quote
    return value


def MakeCSVLine(fields, separator=";", quote='"'):
    formatstring = separator.join([field[0] for field in fields])
    strings = [Quote(field[1], separator, quote) for field in fields]
    return formatstring % tuple(strings)


def ProcessFile(
    filename, options, plugins, list_of_dict, disarmed_filebuffers=None, filebuffer=None
):
    xmlDoc = PDFiD(
        filename,
        options.all,
        options.extra,
        options.disarm,
        options.force,
        disarmed_filebuffers=disarmed_filebuffers,
        filebuffer=filebuffer,
    )
    if plugins == [] and options.select == "":
        PDFID2Dict(xmlDoc, options.nozero, options.force, list_of_dict)
        Print(PDFiD2String(xmlDoc, options.nozero, options.force), options)
        return

    oPDFiD = cPDFiD(xmlDoc, options.force)
    if options.select:
        if options.force or not oPDFiD.errorOccured and oPDFiD.isPDF:
            try:
                selected = eval(options.select)
            except Exception as e:
                Print(
                    "Error evaluating select expression: %s" % options.select, options
                )
                if options.verbose:
                    raise e
                return
            if selected:
                if options.csv:
                    Print(filename, options)
                else:
                    Print(PDFiD2String(xmlDoc, options.nozero, options.force), options)
    else:
        for cPlugin in plugins:
            if not cPlugin.onlyValidPDF or not oPDFiD.errorOccured and oPDFiD.isPDF:
                try:
                    oPlugin = cPlugin(oPDFiD, options.pluginoptions)
                except Exception as e:
                    Print("Error instantiating plugin: %s" % cPlugin.name, options)
                    if options.verbose:
                        raise e
                    return

                try:
                    score = oPlugin.Score()
                except Exception as e:
                    Print("Error running plugin: %s" % cPlugin.name, options)
                    if options.verbose:
                        raise e
                    return

                if options.csv:
                    if score >= options.minimumscore:
                        Print(
                            MakeCSVLine(
                                (
                                    ("%s", filename),
                                    ("%s", cPlugin.name),
                                    ("%.02f", score),
                                )
                            ),
                            options,
                        )
                else:
                    if score >= options.minimumscore:
                        Print(
                            PDFiD2String(xmlDoc, options.nozero, options.force), options
                        )
                        Print(f"{cPlugin.name} score:        {score:.02f}", options)
                        try:
                            Print(
                                f"{cPlugin.name} instructions: {oPlugin.Instructions(score)}",
                                options,
                            )
                        except AttributeError:
                            pass
            else:
                if options.csv:
                    if oPDFiD.errorOccured:
                        Print(
                            MakeCSVLine(
                                (
                                    ("%s", filename),
                                    ("%s", cPlugin.name),
                                    ("%s", "Error occured"),
                                )
                            ),
                            options,
                        )
                    if not oPDFiD.isPDF:
                        Print(
                            MakeCSVLine(
                                (
                                    ("%s", filename),
                                    ("%s", cPlugin.name),
                                    ("%s", "Not a PDF document"),
                                )
                            ),
                            options,
                        )
                else:
                    Print(PDFiD2String(xmlDoc, options.nozero, options.force), options)


def Scan(
    directory,
    options,
    plugins,
    filename_dict,
    disarmed_filebuffer=None,
    filebuffer=None,
):
    try:
        if os.path.isdir(directory):
            for entry in os.listdir(directory):
                Scan(
                    os.path.join(directory, entry),
                    options,
                    plugins,
                    filename_dict,
                    disarmed_filebuffer,
                    filebuffer,
                )
        else:
            ProcessFile(
                directory,
                options,
                plugins,
                filename_dict,
                disarmed_filebuffer,
                filebuffer,
            )
    except Exception as e:
        #        print directory
        logger.error(e)


#        logger.info(sys.exc_info()[2])
#        print traceback.format_exc()


# function derived from: http://blog.9bplus.com/pdfidpy-output-to-json
def PDFiD2JSON(xmlDoc, force):
    # Get Top Layer Data
    errorOccured = xmlDoc.documentElement.getAttribute("ErrorOccured")
    errorMessage = xmlDoc.documentElement.getAttribute("ErrorMessage")
    filename = xmlDoc.documentElement.getAttribute("Filename")
    header = xmlDoc.documentElement.getAttribute("Header")
    isPdf = xmlDoc.documentElement.getAttribute("IsPDF")
    version = xmlDoc.documentElement.getAttribute("Version")
    entropy = xmlDoc.documentElement.getAttribute("Entropy")

    # extra data
    countEof = xmlDoc.documentElement.getAttribute("CountEOF")
    countChatAfterLastEof = xmlDoc.documentElement.getAttribute(
        "CountCharsAfterLastEOF"
    )
    totalEntropy = xmlDoc.documentElement.getAttribute("TotalEntropy")
    streamEntropy = xmlDoc.documentElement.getAttribute("StreamEntropy")
    nonStreamEntropy = xmlDoc.documentElement.getAttribute("NonStreamEntropy")

    keywords = []
    dates = []

    # grab all keywords
    for node in xmlDoc.documentElement.getElementsByTagName("Keywords")[0].childNodes:
        name = node.getAttribute("Name")
        count = int(node.getAttribute("Count"))
        if int(node.getAttribute("HexcodeCount")) > 0:
            hexCount = int(node.getAttribute("HexcodeCount"))
        else:
            hexCount = 0
        keyword = {"count": count, "hexcodecount": hexCount, "name": name}
        keywords.append(keyword)

    # grab all date information
    for node in xmlDoc.documentElement.getElementsByTagName("Dates")[0].childNodes:
        name = node.getAttribute("Name")
        value = node.getAttribute("Value")
        date = {"name": name, "value": value}
        dates.append(date)

    data = {
        "countEof": countEof,
        "countChatAfterLastEof": countChatAfterLastEof,
        "totalEntropy": totalEntropy,
        "streamEntropy": streamEntropy,
        "nonStreamEntropy": nonStreamEntropy,
        "errorOccured": errorOccured,
        "errorMessage": errorMessage,
        "filename": filename,
        "header": header,
        "isPdf": isPdf,
        "version": version,
        "entropy": entropy,
        "keywords": {"keyword": keywords},
        "dates": {"date": dates},
    }
    complete = [{"pdfid": data}]
    result = json.dumps(complete)
    return result


def File2Strings(filename):
    try:
        f = open(filename)
    except:
        return None
    try:
        return list(map(lambda line: line.rstrip("\n"), f.readlines()))
    except:
        return None
    finally:
        f.close()


def ProcessAt(argument):
    if argument.startswith("@"):
        strings = File2Strings(argument[1:])
        if strings == None:
            raise Exception("Error reading %s" % argument)
        else:
            return strings
    else:
        return [argument]


def AddPlugin(cClass):
    global plugins

    plugins.append(cClass)


class cExpandFilenameArguments:
    def __init__(
        self,
        filenames,
        literalfilenames=False,
        recursedir=False,
        checkfilenames=False,
        expressionprefix=None,
    ):
        self.containsUnixShellStyleWildcards = False
        self.warning = False
        self.message = ""
        self.filenameexpressions = []
        self.expressionprefix = expressionprefix
        self.literalfilenames = literalfilenames

        expression = ""
        if len(filenames) == 0:
            self.filenameexpressions = [["", ""]]
        elif literalfilenames:
            self.filenameexpressions = [[filename, ""] for filename in filenames]
        elif recursedir:
            for dirwildcard in filenames:
                if expressionprefix != None and dirwildcard.startswith(
                    expressionprefix
                ):
                    expression = dirwildcard[len(expressionprefix) :]
                else:
                    if dirwildcard.startswith("@"):
                        for filename in ProcessAt(dirwildcard):
                            self.filenameexpressions.append([filename, expression])
                    elif os.path.isfile(dirwildcard):
                        self.filenameexpressions.append([dirwildcard, expression])
                    else:
                        if os.path.isdir(dirwildcard):
                            dirname = dirwildcard
                            basename = "*"
                        else:
                            dirname, basename = os.path.split(dirwildcard)
                            if dirname == "":
                                dirname = "."
                        for path, dirs, files in os.walk(dirname):
                            for filename in fnmatch.filter(files, basename):
                                self.filenameexpressions.append(
                                    [os.path.join(path, filename), expression]
                                )
        else:
            for filename in list(
                collections.OrderedDict.fromkeys(
                    sum(map(self.Glob, sum(map(ProcessAt, filenames), [])), [])
                )
            ):
                if expressionprefix != None and filename.startswith(expressionprefix):
                    expression = filename[len(expressionprefix) :]
                else:
                    self.filenameexpressions.append([filename, expression])
            self.warning = (
                self.containsUnixShellStyleWildcards
                and len(self.filenameexpressions) == 0
            )
            if self.warning:
                self.message = "Your filename argument(s) contain Unix shell-style wildcards, but no files were matched.\nCheck your wildcard patterns or use option literalfilenames if you don't want wildcard pattern matching."
                return
        if self.filenameexpressions == [] and expression != "":
            self.filenameexpressions = [["", expression]]
        if checkfilenames:
            self.CheckIfFilesAreValid()

    def Glob(self, filename):
        if not (
            "?" in filename or "*" in filename or ("[" in filename and "]" in filename)
        ):
            return [filename]
        self.containsUnixShellStyleWildcards = True
        return glob.glob(filename)

    def CheckIfFilesAreValid(self):
        valid = []
        doesnotexist = []
        isnotafile = []
        for filename, expression in self.filenameexpressions:
            hashfile = False
            try:
                hashfile = (
                    FilenameCheckHash(filename, self.literalfilenames)[0] == FCH_DATA
                )
            except:
                pass
            if filename == "" or hashfile:
                valid.append([filename, expression])
            elif not os.path.exists(filename):
                doesnotexist.append(filename)
            elif not os.path.isfile(filename):
                isnotafile.append(filename)
            else:
                valid.append([filename, expression])
        self.filenameexpressions = valid
        if len(doesnotexist) > 0:
            self.warning = True
            self.message += (
                "The following files do not exist and will be skipped: "
                + " ".join(doesnotexist)
                + "\n"
            )
        if len(isnotafile) > 0:
            self.warning = True
            self.message += (
                "The following files are not regular files and will be skipped: "
                + " ".join(isnotafile)
                + "\n"
            )

    def Filenames(self):
        if self.expressionprefix == None:
            return [filename for filename, expression in self.filenameexpressions]
        else:
            return self.filenameexpressions


class cPluginParent:
    onlyValidPDF = True


def LoadPlugins(plugins, verbose):
    if plugins == "":
        return
    scriptPath = GetScriptPath()
    for plugin in sum(map(ProcessAt, plugins.split(",")), []):
        try:
            if not plugin.lower().endswith(".py"):
                plugin += ".py"
            if os.path.dirname(plugin) == "":
                if not os.path.exists(plugin):
                    scriptPlugin = os.path.join(scriptPath, plugin)
                    if os.path.exists(scriptPlugin):
                        plugin = scriptPlugin
            exec(open(plugin).read())
        except Exception as e:
            logger.error("Error loading plugin: %s" % plugin)
            if verbose:
                raise e


def PDFiDMain(filenames, options, filebuffers=None):
    global plugins
    plugins = []
    LoadPlugins(options.plugins, options.verbose)

    if options.csv:
        if plugins != []:
            Print(
                MakeCSVLine(
                    (("%s", "Filename"), ("%s", "Plugin-name"), ("%s", "Score"))
                ),
                options,
            )
        elif options.select != "":
            Print("Filename", options)

    list_of_dict = {"reports": []}

    disarmed_buffers = {"buffers": []}

    if not filebuffers:
        for filename in filenames:
            if options.scan:
                Scan(
                    filename,
                    options,
                    plugins,
                    list_of_dict["reports"],
                    disarmed_buffers["buffers"],
                )
            else:
                ProcessFile(
                    filename,
                    options,
                    plugins,
                    list_of_dict["reports"],
                    disarmed_buffers["buffers"],
                )
    else:
        for i, filebuffer in enumerate(filebuffers):
            ProcessFile(
                filenames[i],
                options,
                plugins,
                list_of_dict["reports"],
                disarmed_buffers["buffers"],
                filebuffer,
            )

    results = ()

    if options.json:
        results += (list_of_dict,)

    if options.return_disarmed_buffer:
        results += (disarmed_buffers,)

    if len(results) == 1:
        results = results[0]

    return results


def GetOTPParser():
    moredesc = """

    Arguments:
    pdf-file and zip-file can be a single file, several files, and/or @file
    @file: run PDFiD on each file listed in the text file specified
    wildcards are supported

    Source code put in the public domain by Didier Stevens, no Copyright
    Use at your own risk
    https://DidierStevens.com"""
    oParser = optparse.OptionParser(
        usage="usage: %prog [options] [pdf-file|zip-file|url|@file] ...\n"
        + __description__
        + moredesc,
        version="%prog " + __version__,
    )
    oParser.add_option(
        "-s",
        "--scan",
        action="store_true",
        default=False,
        help="scan the given directory",
    )
    oParser.add_option(
        "-a", "--all", action="store_true", default=False, help="display all the names"
    )
    oParser.add_option(
        "-e",
        "--extra",
        action="store_true",
        default=False,
        help="display extra data, like dates",
    )
    oParser.add_option(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="force the scan of the file, even without proper %PDF header",
    )
    oParser.add_option(
        "-d",
        "--disarm",
        action="store_true",
        default=False,
        help="disable JavaScript and auto launch",
    )
    oParser.add_option(
        "-p",
        "--plugins",
        type=str,
        default="",
        help="plugins to load (separate plugins with a comma , ; @file supported)",
    )
    oParser.add_option(
        "-c",
        "--csv",
        action="store_true",
        default=False,
        help="output csv data when using plugins",
    )
    oParser.add_option(
        "-m",
        "--minimumscore",
        type=float,
        default=0.0,
        help="minimum score for plugin results output",
    )
    oParser.add_option(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="verbose (will also raise catched exceptions)",
    )
    oParser.add_option(
        "-S", "--select", type=str, default="", help="selection expression"
    )
    oParser.add_option(
        "-n",
        "--nozero",
        action="store_true",
        default=False,
        help="supress output for counts equal to zero",
    )
    oParser.add_option(
        "-o", "--output", type=str, default="", help="output to log file"
    )
    oParser.add_option(
        "--pluginoptions", type=str, default="", help="options for the plugin"
    )
    oParser.add_option(
        "-l",
        "--literalfilenames",
        action="store_true",
        default=False,
        help="take filenames literally, no wildcard matching",
    )
    oParser.add_option(
        "--recursedir",
        action="store_true",
        default=False,
        help="Recurse directories (wildcards and here files (@...) allowed)",
    )
    oParser.add_option(
        "-j",
        "--json",
        action="store_true",
        default=False,
        help="json output, supports only basic analysis",
    )
    oParser.add_option(
        "-b",
        "--return_disarmed_buffer",
        action="store_true",
        default=False,
        help="Returns the disarmed pdf documents as buffer objects.",
    )
    return oParser


def get_fake_options():
    class FakeOptions:
        def __init__(self):
            self.scan = False
            self.all = False
            self.extra = False
            self.force = False
            self.disarm = False
            self.plugins = ""
            self.csv = False
            self.minimumscore = 0.0
            self.verbose = False
            self.select = ""
            self.nozero = False
            self.output = ""
            self.pluginoptions = ""
            self.literalfilenames = False
            self.recursedir = False
            self.json = False
            self.return_disarmed_buffer = False

    fakeoptions = FakeOptions()
    return fakeoptions


def Main():
    oParser = GetOTPParser()
    (options, args) = oParser.parse_args()

    if len(args) == 0:
        if options.disarm:
            logger.info("Option disarm not supported with stdin")
            options.disarm = False
        if options.scan:
            logger.info("Option scan not supported with stdin")
            options.scan = False
        filenames = [""]
    else:
        try:
            oExpandFilenameArguments = cExpandFilenameArguments(
                args, options.literalfilenames, options.recursedir, False
            )
            filenames = oExpandFilenameArguments.Filenames()
            if oExpandFilenameArguments.warning:
                logger.info(oExpandFilenameArguments.message)
        except Exception as e:
            logger.info(e)
            return
    PDFiDMain(filenames, options)


if __name__ == "__main__":
    Main()



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import io

from PIL import Image as ImageP

from govtech_csg_xcg.securefileupload import logger


def check_integrity(file):
    logger.debug("[Validation module] - Starting image integrity check")

    try:
        image = ImageP.open(io.BytesIO(file.content))
        image.verify()
        image.close()
    except Exception as e:
        logger.error(f"[Validation module] - CHECK: Image integrity (1) - FAILED: {e}")
        return False

    try:
        image = ImageP.open(io.BytesIO(file.content))
        image.transpose(ImageP.FLIP_LEFT_RIGHT)
        image.close()
    except Exception as e:
        logger.error(f"[Validation module] - CHECK: Image integrity (2) - FAILED: {e}")
        return False

    logger.debug("[Validation module] - CHECK: Image integrity - PASSED")
    return True


def validate_file(file):
    logger.debug("[Validation module] - Starting image validation")

    file.validation_results.file_integrity_ok = check_integrity(file)
    file.validation_results.file_integrity_check_done = True

    logger.debug("[Validation module] - Validation: Image - DONE")



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from govtech_csg_xcg.securefileupload import logger

try:
    import yara  # noqa: F401
    from quicksand.quicksand import quicksand

except ImportError:
    quicksand = None


def perform_quicksand_scan(file):
    if quicksand is not None:
        logger.debug("[Validation module] - Running Quicksand")

        qs = quicksand(file.content, timeout=18, strings=True)
        qs.process()

        if qs.results["rating"] > 1:
            file.block = True
            file.validation_results.quicksand_result_ok = False
            file.append_block_reason("QS_detection")
            file.validation_results.quicksand_result_detail = qs.results["risk"]
            logger.warning(
                f"{file.basic_information.name} [Validation module] - Blocking file: Quicksand detection"
            )

    else:
        logger.info(
            "[Validation module] - Skipping Quicksand analysis as Quicksand library is not installed"
        )


def validate_file(file):
    logger.debug("[Validation module] - Starting Quicksand")

    perform_quicksand_scan(file)



from django.urls import path

from . import views

urlpatterns = [
    path(
        "test_no_restrictions/", views.test_no_restrictions, name="test-no-restrictions"
    ),
    path("test_quicksand/", views.test_quicksand, name="test-quicksand"),
    path(
        "test_file_size_limit/", views.test_file_size_limit, name="test-file-size-limit"
    ),
    path(
        "test_filename_length_limit/",
        views.test_filename_length_limit,
        name="test-filename-length-limit",
    ),
    path(
        "test_whitelist_custom/",
        views.test_whitelist_custom,
        name="test-whitelist-custom",
    ),
    path("test_sanitization/", views.test_sanitization, name="test-sanitization"),
    path("test_clamav/", views.test_clamav, name="test-clamav"),
    path("test_yara/", views.test_yara, name="test-yara"),
]



import logging

# Create the Logger
logger = logging.getLogger("PDFiD")

if not logger.handlers:
    # Create the Handler for logging data to console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)

    # Create a Formatter for formatting the log messages
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s][XCG][%(name)s]: %(message)s"
    )
    console_handler.setFormatter(formatter)

    # Set the Handler to the Logger
    logger.addHandler(console_handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from .data import whitelists


def fill_missing_configs(partial_config, full_config):
    # This creates a third dict that combines the two,
    # and allows values from partial_config to overwrite
    # full_config for the same keys.
    filled_config = {**full_config, **partial_config}

    if filled_config["whitelist_name"] != "CUSTOM":
        filled_config["whitelist"] = _get_valid_whitelist(
            filled_config["whitelist_name"]
        )

    return filled_config


def _get_valid_whitelist(whitelist_name):
    if whitelist_name == "AUDIO_ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__AUDIO_ALL
    elif whitelist_name == "APPLICATION_ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__APPLICATION_ALL
    elif whitelist_name == "IMAGE_ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__IMAGE_ALL
    elif whitelist_name == "TEXT_ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__TEXT_ALL
    elif whitelist_name == "VIDEO_ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__VIDEO_ALL
    elif whitelist_name == "AUDIO_RESTRICTIVE":
        whitelist = whitelists.WHITELIST_MIME_TYPES__AUDIO_RESTRICTIVE
    elif whitelist_name == "APPLICATION_RESTRICTIVE":
        whitelist = whitelists.WHITELIST_MIME_TYPES__APPLICATION_RESTRICTIVE
    elif whitelist_name == "IMAGE_RESTRICTIVE":
        whitelist = whitelists.WHITELIST_MIME_TYPES__IMAGE_RESTRICTIVE
    elif whitelist_name == "TEXT_RESTRICTIVE":
        whitelist = whitelists.WHITELIST_MIME_TYPES__TEXT_RESTRICTIVE
    elif whitelist_name == "VIDEO_RESTRICTIVE":
        whitelist = whitelists.WHITELIST_MIME_TYPES__VIDEO_RESTRICTIVE
    elif whitelist_name == "ALL":
        whitelist = whitelists.WHITELIST_MIME_TYPES__ALL
    else:  # RESTRICTIVE or other
        whitelist = whitelists.WHITELIST_MIME_TYPES__RESTRICTIVE

    return whitelist



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from django.core.exceptions import ValidationError

from .modules.helper import get_current_request


def xcg_file_validator(value):
    request = get_current_request()

    if request.block_upload:
        raise ValidationError(
            request.upload_errmsg,
            code="dmf",
        )



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import hashlib
import mimetypes
from dataclasses import dataclass, field
from io import BytesIO

import magic

from govtech_csg_xcg.securefileupload import logger


@dataclass
class BasicFileInformation:
    name: str
    size: int
    content_type: str
    content_type_extra: str
    charset: str
    sha256: str


@dataclass
class DetectionResults:
    filename_splits: list = field(default_factory=list)
    extensions: list = field(default_factory=list)
    signature_mime: str = ""
    guessed_mime: str = ""
    yara_matches: list = field(default_factory=list)


@dataclass
class ValidationResults:
    file_size_ok: bool = True
    matching_extension_signature_request_ok: bool = True
    filename_length_ok: bool = True
    extensions_whitelist_ok: bool = True
    request_whitelist_ok: bool = True
    signature_whitelist_ok: bool = True
    yara_rules_ok: bool = True
    quicksand_result_ok: bool = True
    quicksand_result_detail: str = ""
    clamav_result_ok: bool = True
    clamav_result_detail: str = ""

    file_integrity_ok: bool = True
    file_integrity_check_done: bool = True

    malicious: bool = True

    total_points_overall: int = 0
    guessing_scores: dict = field(default_factory=dict)


@dataclass
class PossibleAttacks:
    mime_manipulation: bool = False
    null_byte_injection: bool = False


@dataclass
class SanitizationResults:
    created_random_filename_with_guessed_extension: bool = False
    disarmed_pdf: bool = False


@dataclass
class FileTypes:
    magic_full: str = ""
    magic_mime: str = ""
    mimetypes_type: str = ""
    content_type: str = ""


class File:
    """
    Common class for all files.
    """

    def __init__(self, file):
        logger.debug("[File class] - Initializing file object")
        self._uploaded_file = file
        self._content = b"".join([chunk for chunk in file.chunks()])
        self._block = False
        self._block_reasons = []
        self._name = file.name

        _, _, hash_sha256, _ = self._get_file_hashes()

        guessing_scores = {
            mime_type: 0 for mime_type in list(mimetypes.types_map.values())
        }

        self.basic_information = BasicFileInformation(
            file.name,
            file.size,
            file.content_type,
            file.content_type_extra,
            file.charset,
            hash_sha256,
        )

        self.file_types = FileTypes(
            magic_full=magic.from_buffer(self.content),
            magic_mime=magic.from_buffer(self.content, mime=True),
            mimetypes_type=mimetypes.guess_type(file.name)[0],
            content_type=file.content_type,
        )

        self.validation_results = ValidationResults(guessing_scores=guessing_scores)
        self.attack_results = PossibleAttacks()
        self.detection_results = DetectionResults()
        self.sanitization_results = SanitizationResults()

    def _get_file_hashes(self):
        logger.debug("[File class] - Retrieving file hashes")
        md5_hash = hashlib.md5()
        sha1_hash = hashlib.sha1()
        sha256_hash = hashlib.sha256()
        sha512_hash = hashlib.sha512()

        for chunk in self._uploaded_file.chunks():
            md5_hash.update(chunk)
            sha1_hash.update(chunk)
            sha256_hash.update(chunk)
            sha512_hash.update(chunk)

        hexdigest_md5 = md5_hash.hexdigest()
        hexdigest_sha1 = sha1_hash.hexdigest()
        hexdigest_sha256 = sha256_hash.hexdigest()
        hexdigest_sha512 = sha512_hash.hexdigest()

        logger.debug(f"[File class] - MD5: {hexdigest_md5}")
        logger.debug(f"[File class] - SHA1: {hexdigest_sha1}")
        logger.debug(f"[File class] - SHA256: {hexdigest_sha256}")
        logger.debug(f"[File class] - SHA512: {hexdigest_sha512}")

        return hexdigest_md5, hexdigest_sha1, hexdigest_sha256, hexdigest_sha512

    @property
    def uploaded_file(self):
        return self._uploaded_file

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, new_content: bytes):
        logger.debug("[File class] - Setting new file content")
        self._content = new_content
        self._uploaded_file.file = BytesIO(new_content)

    @property
    def block(self):
        return self._block

    @block.setter
    def block(self, new_block_status: bool):
        logger.debug("[File class] - Setting new block status")
        self._block = new_block_status

    @property
    def block_reasons(self):
        return self._block_reasons

    def append_block_reason(self, block_reason):
        logger.debug("[File class] - Appending new block reason")
        self._block_reasons.append(block_reason)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name
        self.basic_information.name = new_name
        self._uploaded_file.name = new_name



from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render
from django.urls import reverse

from govtech_csg_xcg.securefileupload.decorator import upload_config

from .forms import TestForm


def process_form(request, view_name):
    if request.method == "POST":
        form = TestForm(request.POST, request.FILES)
        if form.is_valid():
            return HttpResponse(request.FILES["uploaded_file"].name)
        else:
            return HttpResponseForbidden("Upload not allowed")
    else:
        form = TestForm()

    return render(
        request, "test_form.html", {"form": form, "view_path": reverse(view_name)}
    )


def test_no_restrictions(request):
    return process_form(request, "test-no-restrictions")


@upload_config(quicksand=True)
def test_quicksand(request):
    return process_form(request, "test-quicksand")


@upload_config(file_size_limit=200)
def test_file_size_limit(request):
    return process_form(request, "test-file-size-limit")


@upload_config(filename_length_limit=50)
def test_filename_length_limit(request):
    return process_form(request, "test-filename-length-limit")


@upload_config(whitelist_name="CUSTOM", whitelist=["text/plain"])
def test_whitelist_custom(request):
    return process_form(request, "test-whitelist-custom")


# Config for sanitization=True has been performed in settings/default.py
# We test sanitization=True + keep_original_filename=True using settings/keep_filename.py
def test_sanitization(request):
    return process_form(request, "test-sanitization")


def test_clamav(request):
    return process_form(request, "test-clamav")


def test_yara(request):
    return process_form(request, "test-yara")



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from govtech_csg_xcg.securefileupload import logger

from . import basic, pdf


def sanitize(file, request):
    """apply all sanitization"""
    logger.debug("[Sanitizer module] - Starting sanitization")

    # Perform basic file sanitization
    basic.sanitize_file(file, request)

    # Get guessed file type
    file_type = file.detection_results.guessed_mime

    # Perform file type specific sanitization
    if file_type == "application/pdf":
        pdf.sanitize_file(file)
    else:
        pass



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import logging
import mimetypes
import time

from django.http import Http404
from django.urls import resolve

from govtech_csg_xcg.securefileupload import logger

from .conf import settings
from .modules import converter, evaluator
from .modules.helper import set_current_request
from .modules.sanitization import sanitizer
from .modules.validation import validator


class FileUploadValidationMiddleware:
    def __init__(self, get_response):
        # One-time configuration and initialization.
        mimetypes.add_type("image/jpeg", ".jfif")
        logging.getLogger("PIL").setLevel(logging.WARNING)
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        set_current_request(request)

        if request.method == "POST" and len(request.FILES) > 0:
            self._monitor_request(request)

            if request.block_upload:
                logger.warning("[Middleware] - Blocking request.")
                # request = self._invalid(request) # TODO whether remove file?

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        set_current_request(None)

        return response

    def _monitor_request(self, request):
        request.block_upload = False
        request.upload_errmsg = ""
        request.middleware_timers = [time.time()]
        request.upload_config = self._extract_single_upload_config(request)

        files = self._extract_files(request)

        for _, file in files.items():
            logger.info(f"[Middleware] - - Analyzing file [{file.name}]")
            self._analyze_file(file, request)
            _, request.block_upload, request.upload_errmsg = self._evaluate_file(
                file, request
            )
            if not file.block and request.upload_config["sanitization"]:
                self._sanitize_file(file, request)

        sanitized_files = converter.build_files(files)
        setattr(request, "_files", sanitized_files)

        self._print_elapsed_time("COMPLETE", request)

    def _sanitize_file(self, file, request):
        sanitizer.sanitize(file, request)
        self._print_elapsed_time("Sanitizer", request)

    def _extract_files(self, request):
        files = converter.request_to_base_file_objects(request.FILES)

        self._print_elapsed_time("Converter", request)
        return files

    def _invalid(self, request):
        conversion = converter.invalid_file(request)

        return conversion

    def _analyze_file(self, file, request):
        validator.validate(file, request.upload_config)
        self._print_elapsed_time("Validator", request)

    def _evaluate_file(self, file, request):
        file, block_upload, upload_errmsg = evaluator.evaluate(file, request)
        self._print_elapsed_time("Evaluator", request)

        return file, block_upload, upload_errmsg

    def _print_elapsed_time(self, processing_step, request):
        curr_time = time.time()
        execution_last_step = (curr_time - request.middleware_timers[-1]) * 1000
        execution_until_now = (curr_time - request.middleware_timers[0]) * 1000

        if processing_step == "COMPLETE":
            logger.debug(
                "[Middleware] - TOTAL execution time: %s sec"
                % (round(execution_until_now / 1000, 3))
            )
        else:
            logger.debug(
                f"[Middleware] - {processing_step} took {round(execution_last_step, 3)} ms - Total: {round(execution_until_now / 1000, 3)} sec"
            )
        request.middleware_timers.append(curr_time)

    def _extract_single_upload_config(self, request):
        path = request.path

        try:
            resolver = resolve(path)
        except Http404:
            return settings.GLOBAL_UPLOAD_CONFIG

        view_func = resolver.func

        if hasattr(view_func, "url_upload_config"):
            upload_config = view_func.url_upload_config
        else:
            upload_config = settings.GLOBAL_UPLOAD_CONFIG

        return upload_config



# Form with just one field to upload a file
from django import forms

from govtech_csg_xcg.securefileupload.validators import xcg_file_validator


class TestForm(forms.Form):
    uploaded_file = forms.FileField(validators=[xcg_file_validator])



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import re

from govtech_csg_xcg.securefileupload import _thread_locals


def find_hex_pattern(needle, haystack):
    return [m.start() for m in re.finditer(needle, haystack)]


def fill_hex_with_zero(content, start_idx, end_idx):
    for i in range(start_idx, end_idx):
        content.pop(i)
        content.insert(i, 0)
    return content


def add_point_to_guessed_file_type(file, mime):
    if mime in file.validation_results.guessing_scores:
        file.validation_results.guessing_scores[mime] += 1
    else:
        file.validation_results.guessing_scores[mime] = 1

    file.validation_results.total_points_overall += 1


def get_current_request():
    """
    Return the request associated with the current thread.
    """
    return getattr(_thread_locals, "request", None)


def set_current_request(request=None):
    """
    Update the request associated with the current thread.
    """
    _thread_locals.request = request



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from io import BytesIO

import clamd

from govtech_csg_xcg.securefileupload import logger


def get_clamAV_results(file_object):
    logger.debug("[Validation module] - Running clamAV check")

    try:
        # Connects to UNIX socket on /var/run/clamav/clamd.ctl
        clam_daemon = clamd.ClamdUnixSocket()

        clamd_res = clam_daemon.instream(BytesIO(file_object.content))
        if clamd_res["stream"][0] == "FOUND":
            file_object.block = True
            file_object.validation_results.clamav_result_ok = False
            file_object.validation_results.clamav_result_detail = clamd_res["stream"][1]
    except Exception:
        logger.error("[Validation module] - clamAV: Cannot connect to clamAV service")


def validate_file(file):
    logger.debug("[Validation module] - Running antivirus check")
    get_clamAV_results(file)



import logging
import os
import re
import unittest

from django.conf import settings
from django.test import SimpleTestCase
from django.urls import reverse
from parameterized import parameterized

logging.getLogger("securefileupload").setLevel(logging.ERROR)
SETTINGS_MODULE_NAME = os.getenv("DJANGO_SETTINGS_MODULE")


@unittest.skipUnless(SETTINGS_MODULE_NAME == "testproject.settings.default", "")
class TestSecureFileUploadFeatures(SimpleTestCase):
    """Unit tests for the various features of the Secure File Upload middleware."""

    @parameterized.expand(
        [
            ("control.txt",),
            ("malicious_file_eicar.com.txt",),
            ("overly_large_file.jpg",),
            (
                "long_file_name_cxuUUGz166w44LA0neLP3QtYQeF5tgeJuawFCSJYNNAIVxp3mjnoZOFuLibg4u5R.txt",
            ),
            ("legit_pdf_file.pdf",),
            # We don't include pdf_file_with_extension_changed.txt because securefileupload will reject files with manipulated extensions even if all other features are turned off.
        ]
    )
    def test_all_payloads_allowed_when_no_restrictions(self, filename):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-no-restrictions"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, 200)

    @parameterized.expand(
        [
            ("control.txt", 200),
            ("malicious_file_eicar.com.txt", 403),
        ]
    )
    def test_quicksand_detects_malicious_file(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-quicksand"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, status_code)

    @parameterized.expand(
        [
            ("control.txt", 200),
            ("overly_large_file.jpg", 403),
        ]
    )
    def test_file_size_limit_rejects_overly_large_file(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-file-size-limit"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, status_code)

    @parameterized.expand(
        [
            ("control.txt", 200),
            (
                "long_file_name_cxuUUGz166w44LA0neLP3QtYQeF5tgeJuawFCSJYNNAIVxp3mjnoZOFuLibg4u5R.txt",
                403,
            ),
        ]
    )
    def test_filename_length_limit_rejects_long_names(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-filename-length-limit"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, status_code)

    @parameterized.expand(
        [
            ("control.txt", 200),
            ("legit_pdf_file.pdf", 403),
            ("pdf_file_with_extension_changed.txt", 403),
        ]
    )
    def test_whitelist_only_accepts_whitelisted_file_types(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-whitelist-custom"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, status_code)

    def test_sanitization_changes_filename_to_uuid(self):
        filepath = settings.BASE_DIR / "testapp/test_files/control.txt"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-sanitization"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.content, b"control.txt")

        uuid_pattern = (
            "^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$"
        )
        filename_without_extension = response.content.decode("utf-8").split(".")[0]
        self.assertTrue(re.match(uuid_pattern, filename_without_extension))


@unittest.skipUnless(SETTINGS_MODULE_NAME == "testproject.settings.keep_filename", "")
class TestSecureFileUploadKeepFilename(SimpleTestCase):
    """This test case specifically tests the keep_original_filename feature.

    As the effective upload configuration is finalized on Django startup, we
    need to test this feature with a separate "settings.py" configuration, since
    the "default" configuration defines keep_original_filename=False.
    """

    def test_sanitization_does_not_change_filename_to_uuid_if_keep_original_filename(
        self,
    ):
        filepath = settings.BASE_DIR / "testapp/test_files/control.txt"
        with open(filepath, "rb") as file:
            response = self.client.post(
                reverse("test-sanitization"), {"uploaded_file": file}
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"control.txt")


@unittest.skipUnless(SETTINGS_MODULE_NAME == "testproject.settings.use_clamav", "")
class TestSecureFileUploadClamavIntegration(SimpleTestCase):
    """This test case specifically tests the clamav feature, which uses the settings/clamav.py settings module."""

    @parameterized.expand(
        [
            ("control.txt", 200),
            ("malicious_file_eicar.com.txt", 403),
        ]
    )
    def test_clamav_flags_malicious_file(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(reverse("test-clamav"), {"uploaded_file": file})
        self.assertEqual(response.status_code, status_code)


@unittest.skipUnless(SETTINGS_MODULE_NAME == "testproject.settings.use_yara", "")
class TestSecureFileUploadYaraIntegration(SimpleTestCase):
    """This test case specifically tests the clamav feature, which uses the settings/clamav.py settings module."""

    @parameterized.expand(
        [
            ("control.txt", 200),
            ("yara_test.txt", 403),
        ]
    )
    def test_clamav_flags_malicious_file(self, filename, status_code):
        filepath = settings.BASE_DIR / f"testapp/test_files/{filename}"
        with open(filepath, "rb") as file:
            response = self.client.post(reverse("test-yara"), {"uploaded_file": file})
        self.assertEqual(response.status_code, status_code)



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from io import BytesIO

from django.core.files.uploadedfile import InMemoryUploadedFile
from django.utils.datastructures import MultiValueDict

from govtech_csg_xcg.securefileupload import File, logger


def request_to_base_file_objects(request_files):
    """extract file information from request files"""
    logger.debug("[Converter module - Basic] - Starting request to File objects")
    file_objects = {}
    for file_key in request_files:
        request_file = request_files[file_key]
        file_objects[file_key] = File.File(request_file)
    return file_objects


def invalid_file(original_request):
    """remove the file from the request object"""
    sanitized_request = original_request
    setattr(sanitized_request, "_files", None)
    return sanitized_request


def build_files(sanitized_file_objects):
    """build request.FILES dict from the customized file object"""
    logger.debug("[Converter module] - Building FILES")
    all_files_MultiValueDict = MultiValueDict()

    for key in sanitized_file_objects:
        sanitized_file_object = sanitized_file_objects[key]
        sanitized_imuf = build_InMemoryUploadedFile(sanitized_file_object)
        all_files_MultiValueDict.appendlist(key, sanitized_imuf)

    return all_files_MultiValueDict


def build_InMemoryUploadedFile(sanitized_file_object):
    """build the individual InMemoryUploadedFile for request.FILES"""
    logger.debug("[Converter module] - Building InMemoryUploadedFile")

    if sanitized_file_object.block:
        # This file has the block flag, then overwrites it with a 1-byte dummy file to keep the upload going.
        file_imuf = InMemoryUploadedFile(
            BytesIO(b" "),
            "file",
            sanitized_file_object.basic_information.name,
            sanitized_file_object.basic_information.content_type,
            1,
            sanitized_file_object.basic_information.charset,
            sanitized_file_object.basic_information.content_type_extra,
        )
        logger.debug(
            f"[Converter module] - File [{sanitized_file_object.name}]'s content is being overwritten with a dummy 1-byte file"
        )

    else:
        file_imuf = InMemoryUploadedFile(
            BytesIO(sanitized_file_object.content),
            "file",
            sanitized_file_object.basic_information.name,
            sanitized_file_object.basic_information.content_type,
            sanitized_file_object.basic_information.size,
            sanitized_file_object.basic_information.charset,
            sanitized_file_object.basic_information.content_type_extra,
        )

    return file_imuf



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import copy
from functools import wraps

from .conf import settings
from .utils import fill_missing_configs


def upload_config(
    quicksand=None,
    file_size_limit=None,
    filename_length_limit=None,
    whitelist_name=None,
    whitelist=None,
):
    """
    Decorator for views that upload options.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapper_view(request, *args, **kwargs):
            return view_func(request, *args, **kwargs)

        url_upload_config = copy.deepcopy(settings.GLOBAL_UPLOAD_CONFIG)

        if quicksand is not None:
            url_upload_config["quicksand"] = quicksand

        if file_size_limit is not None:
            url_upload_config["file_size_limit"] = file_size_limit

        if filename_length_limit is not None:
            url_upload_config["filename_length_limit"] = filename_length_limit

        if whitelist_name is not None:
            url_upload_config["whitelist_name"] = whitelist_name

        if whitelist is not None:
            url_upload_config["whitelist"] = whitelist

        url_upload_config = fill_missing_configs(
            url_upload_config, settings.GLOBAL_UPLOAD_CONFIG
        )

        _wrapper_view.url_upload_config = url_upload_config
        return _wrapper_view

    return decorator



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import hashlib
import logging

from govtech_csg_xcg.securefileupload import logger

from ...vendor.pdfid import pdfid


def sanitize_file(file):
    """sanitize pdf file using pdfid"""
    logger.debug("[Sanitizer module] - Starting application sanitization")

    try:
        options = pdfid.get_fake_options()
        options.disarm = True
        options.return_disarmed_buffer = True
        logging.getLogger("PDFiD").setLevel(logging.WARNING)

        disarmed_pdf_dict = pdfid.PDFiDMain(["pdffile"], options, [file.content])

        # check whether the file content was modified
        md5_hash = hashlib.sha256()
        md5_hash.update(disarmed_pdf_dict["buffers"][0])
        if md5_hash.hexdigest() != file.basic_information.sha256:
            file.sanitization_results.disarmed_pdf = True

        file.content = disarmed_pdf_dict["buffers"][0]

    except Exception as e:
        logger.error(
            f"[Sanitizer module - PDF] - Error sanitizing PDF to generate a disarmed PDF file: {e}"
        )



from .base import *  # noqa: F401, F403

XCG_SECUREFILEUPLOAD_CONFIG = {
    "quicksand": False,
    "file_size_limit": None,
    "filename_length_limit": None,
    "whitelist_name": "ALL",
    "sanitization": True,
    "keep_original_filename": False,
    "clamav": False,
}



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from govtech_csg_xcg.securefileupload import logger


def evaluate(file, request):
    logger.debug("[Evaluator module] - Starting maliciousness evaluation")

    block_upload = request.block_upload
    upload_errmsg = request.upload_errmsg

    #############################
    # Handle validation results #
    #############################

    # Always block if strict validation fails
    # Block if YARA rule matches

    # 1. strict validation - validation
    strict_val_res = [
        file.validation_results.file_size_ok,
        file.validation_results.matching_extension_signature_request_ok,
        file.validation_results.filename_length_ok,
        file.validation_results.extensions_whitelist_ok,
        file.validation_results.request_whitelist_ok,
        file.validation_results.signature_whitelist_ok,
    ]

    # 2. strict validation - attack
    possible_attack_val_res = [
        not file.attack_results.mime_manipulation,
        not file.attack_results.null_byte_injection,
    ]

    strict_val_success = all(strict_val_res)
    possible_attack_val_success = all(possible_attack_val_res)

    if not (strict_val_success and possible_attack_val_success):
        block_upload = True
        file.block = True
        file.validation_results.malicious = True
        file.append_block_reason("strict_eval_failed")

        upload_errmsg = (
            upload_errmsg + " File: [" + file.basic_information.name + "] " + "ERROR: "
        )

        if not file.validation_results.file_size_ok:
            upload_errmsg = upload_errmsg + "File size not match;"

        if not file.validation_results.matching_extension_signature_request_ok:
            upload_errmsg = upload_errmsg + "File extension and signature not match;"

        if not file.validation_results.filename_length_ok:
            upload_errmsg = upload_errmsg + "Filename length not match;"

        if not file.validation_results.extensions_whitelist_ok:
            upload_errmsg = upload_errmsg + "File extensions whitelist not match;"

        if not file.validation_results.request_whitelist_ok:
            upload_errmsg = upload_errmsg + "Request whitelist not match;"

        if not file.validation_results.signature_whitelist_ok:
            upload_errmsg = upload_errmsg + "Signature whitelist not match;"

        if file.attack_results.mime_manipulation:
            upload_errmsg = upload_errmsg + "File with mime_manipulation;"

        if file.attack_results.null_byte_injection:
            upload_errmsg = upload_errmsg + "File with null_byte_injection;"

        logger.warning("[Evaluator module] - Blocking: Strict evaluation FAILED")
    else:
        logger.debug("[Evaluator module] - Strict evaluation PASSED")

    # 2. relaxed YARA validation
    yara_val_success = file.validation_results.yara_rules_ok

    if not yara_val_success:
        block_upload = True
        file.block = True
        file.validation_results.malicious = True
        file.append_block_reason("yara_eval_failed")
        upload_errmsg = f"{upload_errmsg} File: [{file.basic_information.name}]: \
            YARA evaluation FAILED;"
        logger.warning("[Evaluator module] - Blocking: YARA evaluation FAILED")
    else:
        logger.debug("[Evaluator module] - YARA evaluation PASSED")

    # 3. relaxed quicksand validation
    clamav_val_success = file.validation_results.quicksand_result_ok

    if not clamav_val_success:
        block_upload = True
        file.block = True
        file.validation_results.malicious = True
        file.append_block_reason("QS_detection")
        upload_errmsg = f"{upload_errmsg} File: [{file.basic_information.name}]: \
            Quicksand evaluation FAILED: {file.validation_results.quicksand_result_detail};"
        logger.warning("[Evaluator module] - Blocking: Quicksand evaluation FAILED")
    else:
        logger.debug("[Evaluator module] - Quicksand evaluation PASSED")

    # 4. relaxed clamav validation
    clamav_val_success = file.validation_results.clamav_result_ok

    if not clamav_val_success:
        block_upload = True
        file.block = True
        file.validation_results.malicious = True
        file.append_block_reason("clamav")
        upload_errmsg = f"{upload_errmsg} File: [{file.basic_information.name}]: \
            clamAV evaluation FAILED: {file.validation_results.clamav_result_detail};"
        logger.warning("[Evaluator module] - Blocking: clamAV evaluation FAILED")
    else:
        logger.debug("[Evaluator module] - clamAV evaluation PASSED")

    return file, block_upload, upload_errmsg



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import mimetypes
import uuid

from govtech_csg_xcg.securefileupload import logger


def create_random_filename_with_guessed_extension(file):
    """change the file's name to an `uuid4` string"""
    logger.debug("[Sanitizer module] - Creating random file name")
    file_extension = mimetypes.guess_extension(file.detection_results.guessed_mime)
    unique_file_name = str(uuid.uuid4()) + file_extension
    file.name = unique_file_name


def sanitize_file(file, request):
    """apply basic sanitization"""
    logger.debug("[Sanitizer module] - Starting basic sanitization")

    if not request.upload_config["keep_original_filename"]:
        create_random_filename_with_guessed_extension(file)
        file.sanitization_results.created_random_filename_with_guessed_extension = True



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
from govtech_csg_xcg.securefileupload import logger

from . import antivirus, basic, image, quicksand


def validate(file, upload_config):
    logger.debug("[Validation module] - Starting validation")

    # Perform basic file validation
    basic.validate_file(file, upload_config)

    if not file.block:
        file_type = file.detection_results.guessed_mime

        # Perform file type specific validation
        if file_type.startswith("image"):
            image.validate_file(file)

    if not file.block and upload_config["quicksand"]:
        # Perform quicksand scan
        quicksand.validate_file(file)
    if not file.block and upload_config["clamav"]:
        antivirus.validate_file(file)

    logger.debug(
        f"[Validation module] - Current block status: {file.block} => {file.block_reasons}"
    )



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import os
from pathlib import Path

from django.conf import settings

from ..utils import fill_missing_configs

DEFAULT_UPLOAD_CONFIG = {
    "quicksand": False,
    "file_size_limit": None,
    "filename_length_limit": None,
    "whitelist_name": "RESTRICTIVE",
    "whitelist": [],
    "sanitization": True,
    "keep_original_filename": False,
    "clamav": False,
    "yara_file_location": os.path.join(
        Path(os.path.realpath(__file__)).parent.parent.absolute(), "vendor/yara"
    ),
}

GLOBAL_UPLOAD_CONFIG = getattr(
    settings, "XCG_SECUREFILEUPLOAD_CONFIG", DEFAULT_UPLOAD_CONFIG
)
GLOBAL_UPLOAD_CONFIG = fill_missing_configs(GLOBAL_UPLOAD_CONFIG, DEFAULT_UPLOAD_CONFIG)



# ------------------------------------------------------------------------
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# This file incorporates work covered by the following copyright:
#
# Copyright (c) 2024 Agency for Science, Technology and Research (A*STAR).
#   All rights reserved.
# Copyright (c) 2024 Government Technology Agency (GovTech).
#   All rights reserved.
# ------------------------------------------------------------------------
import mimetypes

all_mimetypes = list(mimetypes.types_map.values())

WHITELIST_MIME_TYPES__AUDIO_ALL = [
    mimetype for mimetype in all_mimetypes if mimetype.startswith("audio/")
]
WHITELIST_MIME_TYPES__APPLICATION_ALL = [
    mimetype for mimetype in all_mimetypes if mimetype.startswith("application/")
]
WHITELIST_MIME_TYPES__IMAGE_ALL = [
    mimetype for mimetype in all_mimetypes if mimetype.startswith("image/")
]
WHITELIST_MIME_TYPES__TEXT_ALL = [
    mimetype for mimetype in all_mimetypes if mimetype.startswith("text/")
]
WHITELIST_MIME_TYPES__VIDEO_ALL = [
    mimetype for mimetype in all_mimetypes if mimetype.startswith("video/")
]

WHITELIST_MIME_TYPES__AUDIO_RESTRICTIVE = ["audio/mpeg"]
WHITELIST_MIME_TYPES__APPLICATION_RESTRICTIVE = ["application/pdf"]
WHITELIST_MIME_TYPES__IMAGE_RESTRICTIVE = [
    "image/gif",
    "image/jpeg",
    "image/png",
    "image/tiff",
]
WHITELIST_MIME_TYPES__TEXT_RESTRICTIVE = ["text/plain"]
WHITELIST_MIME_TYPES__VIDEO_RESTRICTIVE = ["video/mp4", "video/mpeg", "video/quicktime"]

WHITELIST_MIME_TYPES__ALL = all_mimetypes
WHITELIST_MIME_TYPES__RESTRICTIVE = (
    WHITELIST_MIME_TYPES__AUDIO_RESTRICTIVE
    + WHITELIST_MIME_TYPES__APPLICATION_RESTRICTIVE
    + WHITELIST_MIME_TYPES__IMAGE_RESTRICTIVE
    + WHITELIST_MIME_TYPES__TEXT_RESTRICTIVE
    + WHITELIST_MIME_TYPES__VIDEO_RESTRICTIVE
)
